/*
 * libmm-streamrecorder
 *
 * Copyright (c) 2000 - 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Contact: Hyuntae Kim <ht1211.kim@samsung.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/*=======================================================================================
|  INCLUDE FILES									|
=======================================================================================*/
#include <stdio.h>
#include <stdarg.h>
#include <sys/vfs.h>			/* struct statfs */

#include "mm_streamrecorder_internal.h"
#include "mm_streamrecorder_util.h"
#include "mm_streamrecorder_fileinfo.h"

/*-----------------------------------------------------------------------
|    GLOBAL VARIABLE DEFINITIONS for internal				|
-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
|    LOCAL VARIABLE DEFINITIONS for internal				|
-----------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
|    LOCAL FUNCTION PROTOTYPES:												|
---------------------------------------------------------------------------*/
/* STATIC INTERNAL FUNCTION */

/*===========================================================================================
|																							|
|  FUNCTION DEFINITIONS																		|
========================================================================================== */
/*---------------------------------------------------------------------------
|    GLOBAL FUNCTION DEFINITIONS:											|
---------------------------------------------------------------------------*/

gboolean _mmstreamrecorder_update_composition_matrix(FILE *f, int orientation)
{
	/* for 0 degree */
	guint32 a = 0x00010000;
	guint32 b = 0;
	guint32 c = 0;
	guint32 d = 0x00010000;

	switch (orientation) {
	case MM_STREAMRECORDER_TAG_VIDEO_ORT_90:	/* 90 degree */
		a = 0;
		b = 0x00010000;
		c = 0xffff0000;
		d = 0;
		break;
	case MM_STREAMRECORDER_TAG_VIDEO_ORT_180:	/* 180 degree */
		a = 0xffff0000;
		d = 0xffff0000;
		break;
	case MM_STREAMRECORDER_TAG_VIDEO_ORT_270:	/* 270 degree */
		a = 0;
		b = 0xffff0000;
		c = 0x00010000;
		d = 0;
		break;
	case MM_STREAMRECORDER_TAG_VIDEO_ORT_NONE:	/* 0 degree */
	default:
		break;
	}

	write_to_32(f, a);
	write_to_32(f, b);
	write_to_32(f, 0);
	write_to_32(f, c);
	write_to_32(f, d);
	write_to_32(f, 0);
	write_to_32(f, 0);
	write_to_32(f, 0);
	write_to_32(f, 0x40000000);

	_mmstreamrec_dbg_log("orientation : %d, write data 0x%x 0x%x 0x%x 0x%x", orientation, a, b, c, d);

	return TRUE;
}

guint64 _mmstreamrecorder_get_container_size(const guchar *size)
{
	guint64 result = 0;
	guint64 temp = 0;

	temp = size[0];
	result = temp << 24;
	temp = size[1];
	result = result | (temp << 16);
	temp = size[2];
	result = result | (temp << 8);
	result = result | size[3];

	_mmstreamrec_dbg_log("result : %lld", (unsigned long long)result);

	return result;
}

gboolean _mmstreamrecorder_update_size(FILE *f, gint64 prev_pos, gint64 curr_pos)
{
	_mmstreamrec_dbg_log("size : %" G_GINT64_FORMAT "", curr_pos - prev_pos);
	if (fseek(f, prev_pos, SEEK_SET) != 0) {
		_mmstreamrec_dbg_err("fseek() fail");
		return FALSE;
	}

	if (!write_to_32(f, curr_pos - prev_pos))
		return FALSE;

	if (fseek(f, curr_pos, SEEK_SET) != 0) {
		_mmstreamrec_dbg_err("fseek() fail");
		return FALSE;
	}

	return TRUE;
}

gboolean _mmstreamrecorder_write_udta_m4a(FILE *f)
{
	gint64 current_pos, pos;

	_mmstreamrec_dbg_log("");

	if ((pos = ftell(f)) < 0) {
		_mmstreamrec_dbg_err("ftell() returns negative value");
		return FALSE;
	}

	if (!write_to_32(f, 0))		/* udta atomic size */
		return FALSE;

	if (!write_tag(f, "udta"))	/* user data fourcc */
		return FALSE;

	if ((current_pos = ftell(f)) < 0) {
		_mmstreamrec_dbg_err("ftell() returns negative value");
		return FALSE;
	}

	if (!_mmstreamrecorder_update_size(f, pos, current_pos))
		return FALSE;

	return TRUE;
}

gboolean _mmstreamrecorder_write_udta(FILE *f, _MMStreamRecorderLocationInfo info)
{
	gint64 current_pos, pos;

	_mmstreamrec_dbg_log("");

	if ((pos = ftell(f)) < 0) {
		_mmstreamrec_dbg_err("ftell() returns negative value");
		return FALSE;
	}

	if (!write_to_32(f, 0))		/*size */
		return FALSE;

	if (!write_tag(f, "udta"))	/* type */
		return FALSE;

	if ((current_pos = ftell(f)) < 0) {
		_mmstreamrec_dbg_err("ftell() returns negative value");
		return FALSE;
	}

	if (!_mmstreamrecorder_update_size(f, pos, current_pos))
		return FALSE;

	return TRUE;
}
